#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================


import bpy
import os
import platform
import inspect
import re
import math
from mathutils import Vector, Matrix, Euler


# Bunch
class Bunch(object):
    pass


def global_property_group(attribute_name):
    """ Retreive PropertyGroup (unique per blend file)
    - registered as "bpy.types.Text.attribute_name"
    - on hidden text block named ".attribute_name"
    """
    name = f'.{attribute_name}'

    if name not in bpy.data.texts:
        text = bpy.data.texts.new(name)
    else:
        text = bpy.data.texts[name]

    if hasattr(text, attribute_name):
        return getattr(text, attribute_name)

    return None


def set_global_property_group(attribute_type):
    """ Set PropertyGroup (to be used with "global_property_group")
    - <=> bpy.types.Text.attribute_type = bpy.props.PointerProperty(name='attribute_type', type=attribute_type)
    """
    attribute_name = f'{attribute_type.__name__}'

    prop = bpy.props.PointerProperty(name=attribute_name, type=attribute_type)

    setattr(bpy.types.Text, attribute_name, prop)


# Module
def method_exists(module, method_name):
    return method_name in [m[0] for m in inspect.getmembers(module)]


# File
def get_filename():
    if bpy.data.filepath is None or bpy.data.filepath == '':
        return None

    return os.path.splitext(bpy.path.basename(bpy.data.filepath))[0]


def get_userdocument_path():

    if is_windows():
        CSIDL_PERSONAL = 5       # My Documents
        SHGFP_TYPE_CURRENT = 0   # Get current, not default value
        
        import ctypes.wintypes

        buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(None, CSIDL_PERSONAL, None, SHGFP_TYPE_CURRENT, buf)
        return conform_path(buf.value + '/') #.replace('\\','/') + '/'

    # Should support Windows, Linux
    return conform_path(os.path.expanduser('~/Documents/'))


def conform_path(path):
    import os
    path = path.replace('\\', os.sep)
    path = path.replace('/', os.sep)
    return path


def make_path_absolute(path):
    fullpath = hasattr(re.match('[a-zA-Z]:', path), 'group')
    bl_relative = path.startswith('//')

    if fullpath or bl_relative:
        return os.path.abspath(bpy.path.abspath(path))

    return path


def is_mac():
    return platform.system().lower() == 'darwin'


def is_windows():
    return platform.system().lower() == 'windows'


def conform_os_app(path):
    if is_mac():
        return f'open {path}'

    return path


# UI
import textwrap
 
def label_multiline(context, layout, text, first_line_red=True, all_line_red=False, icon=''):
    return label_multiline_width(context.region.width, layout, text, first_line_red, all_line_red, icon)


def label_multiline_width(width, layout, text, first_line_red=True, all_line_red=False, icon=''):
    # if not hasattr(context, 'region'):
    #     return

    chars = int(width / 7)   # 7 pix on 1 character
    wrapper = textwrap.TextWrapper(width=chars)
    lines = text.split('\n')

    # first_line_red = True
    first_line = True

    for line in lines:
        text_lines = wrapper.wrap(text=line)
        for text_line in text_lines:
            row = layout.row()
            row.alert = first_line_red or all_line_red
            
            if first_line and icon:
                row.label(text=text_line, icon=icon)
            else:
                row.label(text=text_line)

            first_line_red = False
            first_line = False


# Register/Unregister Keymap
def register(classes):
    for rclass in classes:
        # bpy.utils.register_class(rclass)
        if not hasattr(bpy.types, rclass.__name__):
            try:
                bpy.utils.register_class(rclass)
            except:
                pass



def unregister(classes):
    for rclass in reversed(classes):
        # bpy.utils.unregister_class(rclass)
        if hasattr(bpy.types, rclass.__name__):
            try:
                bpy.utils.unregister_class(rclass)
            except:
                pass


def add_keymap(addon_keymaps, bl_idname, key, shift=False, ctrl=False, alt=False, name='Pose', space_type= 'EMPTY'):
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.new(name=name, space_type=space_type)
        kmi = km.keymap_items.new(bl_idname, type=key, value='PRESS', shift=shift, ctrl=ctrl, alt=alt)
        addon_keymaps.append((km, kmi))


def clean_keymap(addon_keymaps):
    for km,kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()



# Collection
def moveto_collection(obj, col_name, col_parent_name=''):
    c = bpy.data.collections.get(col_name)        
    if c is None:
        c = bpy.data.collections.new(col_name)
                    
        cp = bpy.data.collections.get(col_parent_name)            
        if cp is None:
            cp = bpy.context.scene.collection
            
        if c.name not in cp.children:
            cp.children.link(c)
    
    if obj.name not in c.objects:
        c.objects.link(obj)


# Refresh
def force_update():
    bpy.context.view_layer.update()
    scene = bpy.context.scene
    scene.frame_set(scene.frame_current)


# Maths
to_deg = 180.0 / math.pi
to_rad = math.pi / 180.0

def signed_angle(vector_u, vector_v, normal):
    # Normal specifies orientation
    angle = vector_u.angle(vector_v)
    if vector_u.cross(vector_v).angle(normal) < 1:
        angle = -angle
    return angle


def add_euler(euler_a, euler_b):
    return Euler([euler_a[i] + euler_b[i] for i in range(len(euler_a))])


def substract_euler(euler_a, euler_b):
    return Euler([euler_a[i] - euler_b[i] for i in range(len(euler_a))])


# Keyframes
def insert_keyframe(object):
    dpath = 'location'
    object.keyframe_insert(data_path=dpath)


# Animation Curves
data_pathes = []
def push_fcurve(data_path):
    data_pathes.append(data_path)
    
    
def bash_fcurves():
    obj = bpy.context.object

    if obj.animation_data is None:
        return

    if obj.animation_data.action is None:
        return

    for fcurve in obj.animation_data.action.fcurves:
        if fcurve.data_path in data_pathes:
            
            for kf in fcurve.keyframe_points:
#                kf = fcurve.keyframe_points[-1]
            
                if kf.interpolation != 'CONSTANT':
                    kf.interpolation = 'CONSTANT'
                    
    data_pathes.clear()


def assign_group(grp_name, data_paths):
    obj = bpy.context.object
    action = obj.animation_data.action

    grp = action.groups.get(grp_name)
    if grp is None:
        grp = action.groups.new(grp_name)
        
    for fcurve in obj.animation_data.action.fcurves:
        if fcurve.data_path in data_paths and fcurve.group != grp:
            fcurve.group = grp


# Drivers
def add_driver(obj, property_path, obj_target, data_path, expression='var', var_name='var', bone_target=''):
    fcurves = obj.driver_add(property_path)
    
    from collections.abc import Iterable
    if not isinstance(fcurves, Iterable):
        fcurves = [fcurves]

    for fcurve in fcurves:
        driver = fcurve.driver
        
        var = driver.variables.new()
        var.type = 'SINGLE_PROP'
        var.name = var_name
        target = var.targets[0]
        target.id = obj_target
        target.bone_target = bone_target
        target.data_path = data_path
        driver.expression = expression


# Armature
def get_meshes_from_arma(arma):
    objs = []
    for obj in bpy.data.objects:
        if (obj.type == 'MESH' and
            arma in [m.object for m in obj.modifiers if m.type == 'ARMATURE']):
                objs.append(obj)

    return objs


# MessageBox
def show_message_box(message = '', title = '', icon = 'BLENDER'):

    def draw(self, context):
        self.layout.label(text = message)

    bpy.context.window_manager.popup_menu(draw, title = title, icon = icon)


# Motion Path
def update_motion_path(obj_name):
    current_mode = bpy.context.mode    
    bpy.ops.object.mode_set(mode='OBJECT', toggle=True)

    # act_obj = bpy.context.active_object
    objs = bpy.context.selected_objects
    for obj in objs:
        obj.select_set(False)

    o = bpy.data.objects[obj_name]
    o.select_set(True)
    # bpy.context.view_layer.objects.active = o
    # bpy.context.active_object = o

    bpy.ops.object.paths_update()

    o.select_set(False)
    objs = bpy.context.selected_objects
    for obj in objs:
        obj.select_set(True)
    # bpy.context.active_object = act_obj
    # bpy.context.view_layer.objects.active = act_obj

    bpy.ops.object.mode_set(mode=current_mode)


# ColorSpace
def auto_set_image_colorspace(image, non_color=False, force_colorspace=None):
    image.colorspace_settings.is_data = non_color

    if non_color:
        return

    if force_colorspace is not None:
        try:
            image.colorspace_settings.name = force_colorspace
            return
        except:
            print(f'Colorspace "{force_colorspace}" not found.')

    dd = bpy.context.scene.display_settings.display_device

    cs = {
        'sRGB': ['sRGB'],
        'ACES': ['Utility - sRGB - Texture', 'role_matte_paint'],
    }

    for display, colorspaces in cs.items():
        if display == dd:
            for colorspace in colorspaces:
                try:
                    image.colorspace_settings.name = colorspace
                    return
                except:
                    print(f'Colorspace "{colorspace}" not found for display "{display}"')


# Vertex Group
def delete_vertex_groups(obj):
    for vg in obj.vertex_groups:
        obj.vertex_groups.remove(vg)



# JSON
import json
def save_json(bunch, filepath):
    # Serializing json
    json_object = json.dumps(bunch.__dict__, indent=4)

    if os.path.exists(filepath):
        with open(filepath, 'r') as f:
            file_data = json.load(f)

            # for key, val in file_data.items():
            #     vb = bunch.__dict__.get(key)

            #     if isinstance(val, list) and isinstance(vb, list):
            #         val.extend(vb)
            #         # val = list(set(val))

            #     bunch.__dict__[key] = val

            bunch.__dict__.update(file_data)
            json_object = json.dumps(bunch.__dict__, indent=4)

    # Writing to sample.json
    with open(filepath, 'w') as f:
        f.write(json_object)



#