#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================


import bpy
from bpy.types import BoneGroup, EditBone
from mathutils import Vector, Matrix
from . import tools


# Maths
def compute_pole_location(b0, b1, dist=10):
    head0 = get_bone_worldlocation(b0)
    tail0 = get_bone_worldlocation(b0, 1)
    head1 = get_bone_worldlocation(b1)
    tail1 = get_bone_worldlocation(b1, 1)

    l0 = (tail0 - head0).length
    l1 = (tail1 - head1).length
    lT = l0 + l1

    m = head0 + (tail1 - head0) * l0 / lT
    # m = (head0 * l0 + tail1 * l1) / lT

    angle = b0.y_axis.angle(b1.y_axis)

    u = tail0 - m
    lu = u.length
    # if lu < 1:
    if angle < 0.3:
        u = -b1.z_axis
    else:
        u /= lu
    
    p = tail0 + (u * 0.5 * (l0 + l1)).normalized() * dist * 3
    return p


def get_pole_angle_radian(base_bone, ik_bone, pole_location):
    base_bone_head = get_bone_worldlocation(base_bone)
    base_bone_tail = get_bone_worldlocation(base_bone, 1)
    ik_bone_head = get_bone_worldlocation(ik_bone)
    ik_bone_tail = get_bone_worldlocation(ik_bone, 1)

    pole_normal = (ik_bone_tail - base_bone_head).cross(pole_location - base_bone_head)
    projected_pole_axis = pole_normal.cross(base_bone_tail - base_bone_head)
    return tools.signed_angle(base_bone.x_axis, projected_pole_axis, base_bone_tail - base_bone_head)


def get_pole_angle(base_bone, ik_bone, pole_location):
    pole_angle_in_radians = get_pole_angle_radian(base_bone, ik_bone, pole_location)
    return pole_angle_in_radians * 180 / 3.141592


# Location
def get_bone_worldlocation(pbone, headtail=0):
    return pbone.head.lerp(pbone.tail, headtail)

    arma = pbone.id_data
    y = pbone.length * headtail / pbone.scale.y
    return arma.matrix_world @ pbone.matrix @ Vector((0,y,0))


def set_bone_worldlocation(pbone, location):
    arma = pbone.id_data
    pbone.matrix.translation = arma.matrix_world.inverted_safe() @ location


# Layer
def move_to_layer(pbone, layer):
    pbone.bone.layers[layer] = True

    for i in range(32):
        if i != layer:
            pbone.bone.layers[i] = False
    
    
# Creation
def get_create_bone(arma, head, tail, name, bparent=None, connect=True, layer=0, deform=False):
    pb = arma.pose.bones.get(name)

    if pb is None:
        pb = create_bone(arma, head, tail, name, bparent, connect)

    move_to_layer(pb, layer)
    pb.bone.use_deform = deform

    return pb


def create_bone(arma, head, tail, name, bparent=None, connect=True, deform=False):
    current_mode = bpy.context.mode

    was_hidden = arma.hide_get()
    was_selected = arma.select_get()
    arma.hide_set(False)
    arma.select_set(True)
    
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    
    bone = arma.data.edit_bones.new(name)
    bone.head = head
    bone.tail = tail
    
    if bparent is not None:
        eb_p = arma.data.edit_bones[bparent.name]
        bone.roll = eb_p.roll
        bone.parent = eb_p
        bone.use_connect = connect
    else:
        bone.use_connect = False
    
    bone_name = bone.name
    bpy.ops.object.mode_set(mode=current_mode)

    arma.select_set(was_selected)
    arma.hide_set(was_hidden)
    
    pb = arma.pose.bones[bone_name]
    pb.bone.use_deform = deform
    
    return pb


def link_ctrl_bones(arma_obj, arma_ctrl_obj):
    for pbone in arma_obj.pose.bones:
        add_copy_transform(pbone, arma_ctrl_obj.pose.bones[pbone.name])

    current_mode = bpy.context.mode
    
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    
    for ebone in arma_obj.data.edit_bones:
        ebone.use_connect = False

    bpy.ops.object.mode_set(mode=current_mode)


def create_pole(b0, b1, dist=10):
    arma = b0.id_data    
    current_mode = bpy.context.mode

    p = compute_pole_location(b0, b1, dist)

    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

    bone = arma.data.edit_bones.new(b0.name + '_pole')
    bone.head = p
    bone.tail = bone.head
    bone.tail.z += 0.25 * 0.5 * (b0.length + b1.length)
    bone.roll = 0.0000
    bone.use_connect = False
    
    bone_name = bone.name
    bpy.ops.object.mode_set(mode=current_mode)
    
    pb = arma.pose.bones[bone_name]
    pb.bone.use_deform = False
    return pb


def create_target_ik(b0, suffix, parent=None):
    arma = b0.id_data
    current_mode = bpy.context.mode

    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

    bone = arma.data.edit_bones.new(b0.name + suffix)
    bone.head = b0.head
    bone.tail = bone.head + b0.y_axis * 2.0 * b0.length
    bone.roll = 0.0000
    bone.use_connect = False
    
    if parent is not None:
        bone.parent = arma.data.edit_bones[parent.name]
        
    bone_name = bone.name
    bpy.ops.object.mode_set(mode=current_mode)

    pb = arma.pose.bones[bone_name]
    pb.bone.use_deform = False
    return pb


def create_bone_and_linked(arma_ctrl, arma, head, tail, name, bparent=None, connected=False, deform=False):
    m_b = create_bone(arma_ctrl, head, tail, name, bparent, connected, deform)

    aparent = None
    if bparent is not None:
        aparent = arma.pose.bones[bparent.name]
    b = create_bone(arma, head, tail, name, aparent, connected, deform)
    add_copy_transform(b, m_b)

    return m_b, b


# Change
def change_bone_connect(pbone, connect):
    arma = pbone.id_data    
    current_mode = bpy.context.mode
    
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    
    bone = arma.data.edit_bones[pbone.name]
    bone.use_connect = connect

    bpy.ops.object.mode_set(mode=current_mode)


def change_bone_parent(pbone, pb_parent=None):
    arma = pbone.id_data    
    current_mode = bpy.context.mode

    was_hidden = arma.hide_get()
    was_selected = arma.select_get()
    arma.hide_set(False)
    arma.select_set(True)
    
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    
    bone = arma.data.edit_bones[pbone.name]

    if pb_parent is None:
        bone.parent = None
    else:
        bone.parent = arma.data.edit_bones[pb_parent.name]

    bpy.ops.object.mode_set(mode=current_mode)

    arma.hide_set(was_hidden)
    arma.select_set(was_selected)


# Constraints
def add_ik(bone, base_bone, length, btarget, bpole):
    c = bone.constraints.new('IK')

    c.target = btarget.id_data
    c.subtarget = btarget.name

    c.pole_target = bpole.id_data
    c.pole_subtarget = bpole.name

    p = get_bone_worldlocation(bpole)
    # a = get_pole_angle_radian(base_bone, bone, bpole.matrix.translation)
    a = get_pole_angle_radian(base_bone, bone, p)
    c.pole_angle = a
    
    c.chain_count = length
    return c


def add_child_of(b_child, b_parent):
    c = b_child.constraints.new('CHILD_OF')
    c.target = b_parent.id_data
    c.subtarget = b_parent.name
    
    # save_active = bpy.context.active_object
    # save_active_bone = bpy.context.active_object.data.bones.active

    # bpy.context.view_layer.objects.active = b_child.id_data
    # bpy.context.active_object.data.bones.active = b_child.bone
    # # bpy.ops.constraint.childof_set_inverse(constraint="Child Of", owner='BONE')

    # bpy.context.view_layer.objects.active = save_active
    # bpy.context.active_object.data.bones.active = save_active_bone
    return c
    

def add_stretch_to(b_child, b_target):
    c = b_child.constraints.new('STRETCH_TO')
    c.target = b_target.id_data
    c.subtarget = b_target.name
    return c
    

def add_copy_transform(bone, btarget):
    c = bone.constraints.new('COPY_TRANSFORMS')
    c.target = btarget.id_data
    c.subtarget = btarget.name
    return c
    
    
def add_damped_track(bone, btarget, headtail=0, track_axis='TRACK_Y'):
    c = bone.constraints.new('DAMPED_TRACK')
    c.target = btarget.id_data
    c.subtarget = btarget.name
    c.track_axis = track_axis
    c.head_tail = headtail
    return c
    
    
def add_limit_rot(bone, minX, maxX):
    c = bone.constraints.new('LIMIT_ROTATION')
    c.use_limit_x = True
    c.use_limit_y = True
    c.use_limit_z = True
    c.use_transform_limit = True
    
    c.min_x = minX
    c.max_x = maxX * 3.14159 / 180.0
    c.min_y = 0
    c.max_y = 0
    c.min_z = 0
    c.max_z = 0
    
    c.owner_space = 'LOCAL'
    return c
    
    
def add_limit_scale(bone, minX=None, minY=None, minZ=None, maxX=None, maxY=None, maxZ=None, affect_transform=True):
    c = bone.constraints.new('LIMIT_SCALE')

    if minX is not None:
        c.use_min_x = True
        c.min_x = minX

    if minY is not None:
        c.use_min_y = True
        c.min_y = minX

    if minZ is not None:
        c.use_min_z = True
        c.min_z = minX

    if maxX is not None:
        c.use_max_x = True
        c.max_x = minX

    if maxY is not None:
        c.use_max_y = True
        c.max_y = minX

    if maxZ is not None:
        c.use_max_z = True
        c.max_z = minX

    c.use_transform_limit = affect_transform

    return c
    
    
def add_copy_rotation(bone, btarget, influence=1):
    c = bone.constraints.new('COPY_ROTATION')
    c.target = btarget.id_data
    c.subtarget = btarget.name
    c.influence = influence
    return c


def add_copy_location(bone, btarget, influence=1):
    c = bone.constraints.new('COPY_LOCATION')
    c.target = btarget.id_data
    c.subtarget = btarget.name
    c.influence = influence
    return c


# Selection
restore_stack = []
def push_last():
    last_sel = []
    for pb in bpy.context.selected_pose_bones:
        last_sel.append(pb.name)
    
    if bpy.context.active_object.data.bones.active is None:
        last_active = ''
    else:
        last_active = bpy.context.active_object.data.bones.active.name
    
    restore_stack.append([last_active, last_sel])


def remove_from_last(pbones):
    last_active, last_sel = restore_stack.pop()

    pbone_names = [pb.name for pb in pbones]

    last_sel = [name for name in last_sel if name not in pbone_names]

    restore_stack.append([last_active, last_sel])

    return last_active


def last_active(last, pb, new_pb):
    if last == pb.name:
        if new_pb is None:
            return ''
        return new_pb.name
    return last


def add_to_last(last, sels):
    _, last_sel = restore_stack.pop()

    last_sel.extend(sels)

    restore_stack.append([last, last_sel])


def restore_last():
    last_active, last_sel = restore_stack.pop()
    return restore_all(last_sel, last_active)


def restore_all(last_sel, last_active):
    arma = bpy.context.object
    for pb in bpy.context.selected_pose_bones:
        pb.bone.select = False
        
    for pbname in last_sel:
        arma.pose.bones[pbname].bone.select = True
        
    last_pb = None
    if last_active == '' or last_active not in last_sel:
        bpy.context.active_object.data.bones.active = None
    else:
        last_pb = arma.pose.bones[last_active]
        bpy.context.active_object.data.bones.active = last_pb.bone

    return last_pb


# Post Constraints
def post_constraints():
    for obj in bpy.data.objects:
        if obj.type == 'ARMATURE':
            apply_post_constraints(obj)


def apply_post_constraints(arma):
    l = arma.get('post_constraints')
    if l is None:
        l = []
    else:
        l = list(l)

    for c in l:
        type, pb_dest_name, pb_target_name, condition_bone, condition, value, headtail, headtail1 = c.split(',')

        if type == 'LOCATION':
            post_copy_location(arma, pb_dest_name, pb_target_name, float(headtail), condition_bone, condition, value)
        elif type == 'TRANSFORM':
            post_copy_transform(arma, pb_dest_name, pb_target_name, condition_bone, condition, value)
        elif type == 'DAMPED_TRACK':
            post_damped_track(arma, pb_dest_name, pb_target_name, float(headtail), condition_bone, condition, value)
        elif type == 'DAMPED_LOCATION':
            post_damped_location(arma, pb_dest_name, pb_target_name, condition_bone, condition, value, float(headtail), float(headtail1))
        elif type == 'ZERO_ROT_X':
            post_zero_rot(arma, pb_dest_name, pb_target_name, condition_bone, condition, value, 'X')
        elif type == 'ZERO_ROT_Y':
            post_zero_rot(arma, pb_dest_name, pb_target_name, condition_bone, condition, value, 'Y')
        elif type == 'ZERO_ROT_Z':
            post_zero_rot(arma, pb_dest_name, pb_target_name, condition_bone, condition, value, 'Z')
            

def add_post_constraint(type, pb_dest, pb_target, condition_bone='', condition_key='', condition_value='', headtail=0, headtail1=1):
    arma = pb_dest.id_data
    l = arma.get('post_constraints')
    if l is None:
        l = []
    else:
        l = list(l)
        
    l.append(",".join([type, pb_dest.name, pb_target.name, condition_bone, condition_key, condition_value, str(headtail), str(headtail1)]))
    arma['post_constraints'] = l


def check_condition(obj, condition_bone='', condition_name=None, condition_value=0):  
    if condition_name is None or condition_name == '':
        return True
    
    if condition_bone == '':
        c = obj.get(condition_name)
    else:
        c = obj.pose.bones[condition_bone][condition_name]

    if c is None:
        return False

    return str(c) == str(condition_value)


def post_copy_location(arma, dest_name, target_name, headtail=0, condition_bone='', condition_name=None, condition_value=1):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb0 = arma.pose.bones[target_name]
    pb1 = arma.pose.bones[dest_name]
    
    p0 = get_bone_worldlocation(pb0, headtail)
    set_bone_worldlocation(pb1, p0)
    
    
def post_copy_transform(arma, dest_name, target_name, condition_bone='', condition_name=None, condition_value=1):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb0 = arma.pose.bones[target_name]
    pb1 = arma.pose.bones[dest_name]

    pb1.matrix = pb0.matrix.copy()    
    
    
def post_copy_rotation(arma, dest_name, target_name, condition_bone='', condition_name=None, condition_value=1):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb0 = arma.pose.bones[target_name]
    pb1 = arma.pose.bones[dest_name]
    
    tr = pb1.matrix.translation.copy()
    
    m = pb0.matrix.copy()
    m.translation = tr

    pb1.matrix = m
    
    
def post_damped_track(arma, dest_name, target_name, headtail=0, condition_bone='', condition_name=None, condition_value=1):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb0 = arma.pose.bones[target_name]
    pb1 = arma.pose.bones[dest_name]

    p0 = get_bone_worldlocation(pb0, headtail)
    p1 = get_bone_worldlocation(pb1)

    dir = p0 - p1
    rot_quat = dir.to_track_quat('Y', 'Z')

    m = rot_quat.to_matrix().to_4x4()
    m.translation = arma.matrix_world.inverted_safe() @ p1
    pb1.matrix = m
    
    
def post_damped_location(arma, dest_name, target_name, condition_bone='', condition_name=None, condition_value=1, headtail0=0, headtail1=1):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb0 = arma.pose.bones[target_name]
    pb1 = arma.pose.bones[dest_name]

    p0 = get_bone_worldlocation(pb0, headtail1)
    p1 = get_bone_worldlocation(pb0, headtail0)

    dir = p0 - p1
    rot_quat = dir.to_track_quat('Y', 'Z')

    m = rot_quat.to_matrix().to_4x4()
    m.translation = arma.matrix_world.inverted_safe() @ p1
    pb1.matrix = m
    
    
def post_zero_rot(arma, dest_name, target_name, condition_bone='', condition_name=None, condition_value=1, axe='X'):
    if not check_condition(arma, condition_bone, condition_name, condition_value):
        return
    
    pb = arma.pose.bones[dest_name]
    rot = pb.rotation_quaternion
    if axe == 'X':
        rot.x = 0
    elif axe == 'Y':
        rot.y = 0
    elif axe == 'Z':
        rot.z = 0


# Keyframes
def insert_keyframe_bone(pbone):
    arma = pbone.id_data
    dpath = 'pose.bones["' + pbone.name + '"].'
    dpl = dpath + 'location'
    dpr = dpath + 'rotation_quaternion'
    dpre = dpath + 'rotation_euler'
    dps = dpath + 'scale'
    arma.keyframe_insert(data_path=dpl)
    arma.keyframe_insert(data_path=dpr)
    arma.keyframe_insert(data_path=dpre)
    arma.keyframe_insert(data_path=dps)

    grp_name = get_group_name([pbone.name])
    tools.assign_group(grp_name, [dpl, dpr, dpre, dps])


def insert_keyframe_bone_location(pbone):
    arma = pbone.id_data
    dpath = 'pose.bones["' + pbone.name + '"].'
    dpl = dpath + 'location'
    # dpr = dpath + 'rotation_quaternion'
    # dps = dpath + 'scale'
    arma.keyframe_insert(data_path=dpl)
    # arma.keyframe_insert(data_path=dpr)
    # arma.keyframe_insert(data_path=dps)

    grp_name = get_group_name([pbone.name])
    tools.assign_group(grp_name, [dpl])
    # tools.assign_group(grp_name, [dpl, dpr, dps])


def get_group_name(bone_name_list):
    bone_name = bone_name_list[0]
    
    if bone_name == "thigh_r" or bone_name == "ball_r":
        return "leg_r"
    
    if bone_name == "thigh_l" or bone_name == "ball_l":
        return "leg_l"
    
    if bone_name == "upperarm_r" or bone_name == "clavicle_r":
        return "arm_r"
    
    if bone_name == "upperarm_l" or bone_name == "clavicle_l":
        return "arm_l"
    
    for finger in ['thumb', 'index', 'middle', 'ring', 'pinky']:
        if finger in bone_name:
            if '_r' in bone_name:
                return 'hand_r'
            else:
                return 'hand_l'
    
    if "spine" in bone_name:
        return 'spine'

    for other in ['neck', 'head']:
        if other in bone_name:
            return 'head'
    return bone_name


# Bone orientation
def fix_bone_orientation(pb_bone):
    prim, dir = get_orientation(pb_bone)

    if prim is None or dir is None:
        return

    fix_orientation(pb_bone, prim, dir)


def fix_orientation(pb_bone, prim_axis, positive_dir):
    arma = pb_bone.id_data

    bpy.ops.object.mode_set(mode='EDIT')

    eb = arma.data.edit_bones[pb_bone.name]

    tries = 0

    while tries < 4:

        xp = abs(Vector.dot(eb.x_axis, prim_axis))
        zp = abs(Vector.dot(eb.z_axis, prim_axis))

        xd = Vector.dot(eb.x_axis, positive_dir)
        zd = Vector.dot(eb.z_axis, positive_dir)

        if xp > zp and zd > 0 and zd > xd:
            break

        tries += 1
        eb.roll += 3.14159 * 0.5

    bpy.ops.object.mode_set(mode='POSE')


def get_orientation(pb_bone):
    arma = pb_bone.id_data
    bone_name = pb_bone.name

    ref = arma.pose.bones[0]
    x_axis = ref.x_axis
    y_axis = ref.y_axis
    z_axis = ref.z_axis

    for finger in ['thumb']:
        if finger in bone_name:

            pb_hand = pb_bone
            while finger in pb_hand.name or 'palm' in pb_hand.name:
                pb_hand = pb_hand.parent

            return pb_hand.x_axis, -z_axis

    for finger in ['index', 'middle', 'ring', 'pinky']:
        if finger in bone_name:
            return z_axis, -y_axis

    for arm in ['arm', 'hand', 'clavicle', 'shoulder']:
        if arm in bone_name:
            return y_axis, z_axis

    for spine in ['pelvis', 'hips', 'spine', 'neck', 'head', 'face']:
        if spine in bone_name:
            return x_axis, z_axis

    for leg in ['leg', 'thigh', 'calf', 'foot']:
        if leg in bone_name:
            return x_axis, -z_axis

    return None, None


# Group bones
def assign_bone_group(pb_bone, ref_bone=None):
    arma = pb_bone.id_data

    if ref_bone is not None:
        pb_bone.bone_group = ref_bone.bone_group

    name = pb_bone.name

    bg = None
    if name.endswith('_l'):
        bg = get_bone_group(arma, "left", (0.000,0.400,0.255), (0.000,0.859,0.478), (0.749,1.000,0.910))

    if name.endswith('_r'):
        bg = get_bone_group(arma, "right", (0.506,0.000,0.000), (1.000,0.540,0.314), (1.000,1.000,0.698))

    if bg is not None:
        pb_bone.bone_group = bg


def get_bone_group(arma, name, color_normal, color_select, color_active):
    bone_group = arma.pose.bone_groups.get(name)

    if bone_group is None:
        bone_group = arma.pose.bone_groups.new(name=name)

    bone_group.color_set = "CUSTOM"
    bone_group.colors.normal = color_normal
    bone_group.colors.select = color_select
    bone_group.colors.active = color_active

    return bone_group