import bpy
from ctypes import *

# Handler type enum. Operator is 3
WM_HANDLER_TYPE_GIZMO = 1
WM_HANDLER_TYPE_UI = 2
WM_HANDLER_TYPE_OP = 3
WM_HANDLER_TYPE_DROPBOX = 4
WM_HANDLER_TYPE_KEYMAP = 5

version = bpy.app.version
functype = type(lambda: None)
as_ptr = bpy.types.bpy_struct.as_pointer


class ListBase(Structure):
    _cache = {}
    _fields_ = (("first", c_void_p), ("last",  c_void_p))
    def __new__(cls, c_type=None):
        if c_type in cls._cache: return cls._cache[c_type]
        elif c_type is None: ListBase_ = cls
        else:
            class ListBase_(Structure):
                __name__ = __qualname__ = f"ListBase{cls.__qualname__}"
                _fields_ = (("first", POINTER(c_type)),
                            ("last",  POINTER(c_type)))
                __iter__ = cls.__iter__
                __bool__ = cls.__bool__
        return cls._cache.setdefault(c_type, ListBase_)

    def __iter__(self):
        links_p = []
        elem_n = self.first or self.last
        elem_p = elem_n and elem_n.contents.prev
        if elem_p:
            while elem_p:
                links_p.append(elem_p.contents)
                elem_p = elem_p.contents.prev
            yield from reversed(links_p)
        while elem_n:
            yield elem_n.contents
            elem_n = elem_n.contents.next

    def __bool__(self):
        return bool(self.first or self.last)


class StructBase(Structure):
    _structs = []
    __annotations__ = {}

    def __init_subclass__(cls):
        cls._structs.append(cls)

    def __new__(cls, srna=None):
        if srna is None:
            return super().__new__(cls)
        try:
            return cls.from_address(as_ptr(srna))
        except AttributeError:
            raise Exception("Not a StructRNA instance")

    # Required
    def __init__(self, *_):
        pass


# source\blender\windowmanager\wm_event_system.h
class wmEventHandler(StructBase):
    next:   lambda: POINTER(wmEventHandler)
    prev:   lambda: POINTER(wmEventHandler)
    type:   c_int
    flag:   c_char
    poll:   c_void_p


# source\blender\makesdna\DNA_windowmanager_types.h
class wmWindow(StructBase):
    next:                   lambda: POINTER(wmWindow)
    prev:                   lambda: POINTER(wmWindow)
    ghostwin:               c_void_p
    gpuctx:                 c_void_p
    parent:                 lambda: POINTER(wmWindow)
    scene:                  c_void_p
    new_scene:              c_void_p
    view_layer_name:        c_char * 64

    if version >= (3, 3):
        unpinned_scene:     c_void_p        # Scene

    workspace_hook:         c_void_p
    global_areas:           ListBase * 3    # ScrAreaMap
    screen:                 c_void_p        # bScreen (deprecated)

    if version >= (2, 93):
        winid:              c_int

    pos:                    c_short * 2
    size:                   c_short * 2
    windowstate:            c_char
    active:                 c_char

    if version <= (2, 93):
        _pad0:              c_char * 4

    cursor:                 c_short
    lastcursor:             c_short
    modalcursor:            c_short
    grabcursor:             c_short
    addmousemove:           c_char
    tag_cursor_refresh:     c_char

    if version > (2, 93):
        event_queue_check_click:        c_char
        event_queue_check_drag:         c_char
        event_queue_check_drag_handled: c_char
        _pad0:                          c_char * 1

    else:
        winid:              c_int

    pie_event_type_lock:    c_short
    pie_event_type_last:    c_short
    eventstate:             c_void_p  # wmEvent

    if version >= (3, 2):
        event_last_handled: c_void_p  # wmEvent

    else:
        tweak:              c_void_p # wmGesture

    ime_data:               c_void_p
    event_queue:            ListBase
    handlers:               ListBase(wmEventHandler)
    modalhandlers:          ListBase(wmEventHandler)
    gesture:                ListBase
    stereo3d_format:        c_void_p
    drawcalls:              ListBase
    cursor_keymap_status:   c_void_p


# source\blender\windowmanager\wm_event_system.h
class wmEventHandler_Op(StructBase):

    class context(StructBase):  # Anonymous
        win:            lambda: POINTER(wmWindow)
        area:           c_void_p        # ScrArea ptr
        region:         c_void_p        # ARegion ptr
        region_type:    c_short

    head:               wmEventHandler
    op:                 c_void_p        # wmOperator
    is_file_select:     c_bool
    context:            context

    del context


def init_structs():
    for struct in StructBase._structs:
        fields = []
        anons = []
        for key, value in struct.__annotations__.items():
            if isinstance(value, functype):
                value = value()
            elif isinstance(value, Union):
                anons.append(key)
            fields.append((key, value))

        if anons:
            struct._anonynous_ = anons

        # Base classes might not have _fields_. Don't set anything.
        if fields:
            struct._fields_ = fields
        struct.__annotations__.clear()

    StructBase._structs.clear()
    ListBase._cache.clear()


def is_running_modal(window=None):
    if not window:
        window = bpy.context.window

    win = wmWindow(window)

    for handle in win.modalhandlers:
        if handle.type == WM_HANDLER_TYPE_OP:
            return True
    else:
        return False
        

def register():

# if __name__ == "__main__":
    
    # Important. This sets up the struct fields.
    init_structs()

    # win = wmWindow(bpy.context.window)

    # for handle in win.modalhandlers:
    #     if handle.type == WM_HANDLER_TYPE_OP:
    #         print("Modal running")
    #         break
    # else:
    #     print("No running modals")


def unregister():
    pass