#====================== BEGIN GPL LICENSE BLOCK ======================
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#======================= END GPL LICENSE BLOCK ========================


import bpy
from bpy_extras import view3d_utils
from .libs import tools
from mathutils import Vector



class PHYXDRAG_OT_simulate(bpy.types.Operator):
    bl_idname = "phyxdrag.simulate"
    bl_label = "Simulate"
    bl_options = {'REGISTER', 'UNDO'}

    poll_context = ('OBJECT')


    @classmethod
    def poll(cls, context):
        return context.mode in cls.poll_context


    def get_objects(self):
        return [o for o in bpy.context.scene.objects if o.type == 'MESH']


    def get_selected_objects(self):
        return self.selected_constrained


    def add_to_dict(self, o1, o2, d):
        if o1 in d:
            if o2 not in d[o1]:
                d[o1].append(o2)
        else:
            d[o1] = [o2]


    def create_links(self):
        self.links = dict()
        self.constraints = dict()

        if not bpy.context.scene.rigidbody_world.constraints:
            return

        for obj in bpy.context.scene.rigidbody_world.constraints.all_objects:
            o1 = obj.rigid_body_constraint.object1
            o2 = obj.rigid_body_constraint.object2

            self.add_to_dict(o1, o2, self.links)
            self.add_to_dict(o2, o1, self.links)

            self.add_to_dict(o1, obj, self.constraints)
            self.add_to_dict(o2, obj, self.constraints)

    
    def refresh_constraint_rec(self, obj=None):
        if not hasattr(self, 'selected_constrained'):
            self.create_links()
            self.selected_constrained = []

        if obj is None:
            for o in [o for o in bpy.context.selected_objects if o.type == 'MESH']:
                self.refresh_constraint_rec(o)
    
            self.selected_constrained = list(set(self.selected_constrained))
            return

        cs = self.selected_constrained

        if obj in cs:
            return

        cs.append(obj)

        if obj not in self.links:
            return
        
        for o in self.links[obj]:
            if o in cs:
                continue

            self.refresh_constraint_rec(o)


    def turn_constraints_to_fixed(self):
        self.saved_cs = dict()

        cs = list(set([oc for o in self.selected_constrained if o in self.constraints for oc in self.constraints[o]]))

        for oc in cs:
            c = dict()
            for att in [att for att in dir(oc.rigid_body_constraint) if not att.startswith('__') and att not in ['bl_rna', 'rna_type']]:
                c[att] = getattr(oc.rigid_body_constraint, att)

            self.saved_cs[oc] = c
            oc.rigid_body_constraint.type = 'FIXED'


    def restore_constraints(self):
        for obj, constraint in self.saved_cs.items():
            if obj.rigid_body_constraint != None:
                for att, val in constraint.items():
                    setattr(obj.rigid_body_constraint, att, val)


    def init(self):
        scene = bpy.context.scene
        if scene.rigidbody_world == None:
            bpy.ops.rigidbody.world_add()
            
        scene.rigidbody_world.enabled = True

        if scene.rigidbody_world.collection == None:
            scene.rigidbody_world.collection = bpy.data.collections.new("RigidBodyWorld")

        rigidbody_world = scene.rigidbody_world

        self.saves = dict()
        self.saves['frame_start'] = rigidbody_world.point_cache.frame_start
        self.saves['frame_end'] = rigidbody_world.point_cache.frame_end
        self.saves['substeps_per_frame'] = rigidbody_world.substeps_per_frame
        self.saves['solver_iterations'] = rigidbody_world.solver_iterations
        self.saves['time_scale'] = rigidbody_world.time_scale
        self.saves['effector_weights.gravity'] = rigidbody_world.effector_weights.gravity
        self.saves['effector_weights.all'] = rigidbody_world.effector_weights.all

        self.saves['use_gravity'] = scene.use_gravity

        rigidbody_world.point_cache.frame_start = scene.frame_current
        rigidbody_world.point_cache.frame_end = scene.frame_current + 1
        rigidbody_world.substeps_per_frame = scene.PhyXDragGen.substeps_per_frame
        rigidbody_world.solver_iterations = scene.PhyXDragGen.solver_iterations
        rigidbody_world.time_scale = 1
        rigidbody_world.effector_weights.gravity = 1
        rigidbody_world.effector_weights.all = 1

        scene.use_gravity = True

        self.refresh_constraint_rec()
        self.turn_constraints_to_fixed()

        bpy.context.window.cursor_modal_set('SCROLL_XY')

        if bpy.context.object not in bpy.context.selected_objects:
            bpy.context.view_layer.objects.active = bpy.context.selected_objects[0]
        self.obj_pos = bpy.context.object.location.copy()
        objs = self.get_objects()

        self.save_rigid_by_obj = dict()
        for obj in objs:
            self.save_rigid_by_obj[obj] = None
            if obj.rigid_body == None:
                continue

            rigid = dict()
            for att in [att for att in dir(obj.rigid_body) if not att.startswith('__') and att not in ['bl_rna', 'rna_type']]:
                rigid[att] = getattr(obj.rigid_body, att)

            self.save_rigid_by_obj[obj] = rigid

            # bpy.ops.rigidbody.object_remove({'object': obj})

        self.save_mute_by_gp_by_obj = dict()
        for obj in objs:
            if not obj.animation_data or not obj.animation_data.action:
                continue

            self.save_mute_by_gp_by_obj[obj] = dict()

            for gp in obj.animation_data.action.groups:
                self.save_mute_by_gp_by_obj[obj][gp] = gp.mute
                gp.mute = True

        for obj in [o for o in scene.objects if hasattr(o, 'PhyXDrag') and o.type == 'MESH']:
            selected = obj in self.get_selected_objects()
            participate =  selected or obj.PhyXDrag.Collide

            if not participate:
                continue

            if not obj.rigid_body:
                scene.rigidbody_world.collection.objects.link(obj)

            obj.rigid_body.type = 'ACTIVE' if selected else 'PASSIVE'
            obj.rigid_body.collision_shape = obj.PhyXDrag.Shape
            obj.rigid_body.collision_margin = scene.PhyXDragGen.collision_margin

        self.charge_obj = bpy.data.objects.new("ChargeFieldEmpty", None)
        bpy.context.collection.objects.link(self.charge_obj)

        saved_selection = bpy.context.selected_objects
        saved_active = bpy.context.view_layer.objects.active
        self.charge_obj.select_set(True)
        bpy.context.view_layer.objects.active = self.charge_obj
        bpy.ops.object.forcefield_toggle()
        self.charge_obj.select_set(False)

        for obj in saved_selection:
            obj.select_set(True)
        bpy.context.view_layer.objects.active = saved_active

        self.charge_obj.field.type = 'CHARGE'
        self.charge_obj.field.strength = 10


    def end(self, context, do_undo=False):
        bpy.context.window.cursor_modal_restore()

        # context.window_manager.event_timer_remove(self._timer)

        if do_undo:
            bpy.app.timers.register(lambda : (bpy.ops.ed.undo(), None)[1], first_interval=0.001)

        rigidbody_world = context.scene.rigidbody_world
        rigidbody_world.point_cache.frame_start = self.saves['frame_start']
        rigidbody_world.point_cache.frame_end = self.saves['frame_end']
        rigidbody_world.substeps_per_frame = self.saves['substeps_per_frame']
        rigidbody_world.solver_iterations = self.saves['solver_iterations']
        rigidbody_world.time_scale = self.saves['time_scale']
        rigidbody_world.effector_weights.gravity = self.saves['effector_weights.gravity']
        rigidbody_world.effector_weights.all = self.saves['effector_weights.all']

        context.scene.use_gravity = self.saves['use_gravity']

        for obj, rigid in self.save_rigid_by_obj.items():
            if obj.rigid_body != None and rigid == None:
                ctx = bpy.context.copy()
                ctx['object'] = obj
                with bpy.context.temp_override(**ctx):
                    bpy.ops.rigidbody.object_remove() #{'object': obj})
                # rigidbody_world.collection.objects.unlink(obj)

            elif obj.rigid_body == None and rigid != None:
                rigidbody_world.collection.objects.link(obj)

            if obj.rigid_body != None:
                for att, val in rigid.items():
                    setattr(obj.rigid_body, att, val)

        for obj, mute_by_gp in self.save_mute_by_gp_by_obj.items():
            for gp, mute in mute_by_gp.items():
                gp.mute = mute

        self.restore_constraints()

        bpy.context.collection.objects.unlink(self.charge_obj)
        bpy.data.objects.remove(self.charge_obj, do_unlink=True)



    def modal(self, context, event):
        if event.type in ['ESC','RIGHTMOUSE']:
            self.end(context, True)
            return {'FINISHED'} # if 'CANCELLED' does not trigger auto Undo

        elif event.type in ['LEFTMOUSE']:
            self.end(context)
            return {'FINISHED'}
        
        if event.type == 'MOUSEMOVE':
            try:
                mouse_pos = [event.mouse_region_x, event.mouse_region_y]
                object = bpy.context.object
                region = bpy.context.region
                region3D = bpy.context.space_data.region_3d

                pos = view3d_utils.region_2d_to_location_3d(region, region3D, mouse_pos, self.obj_pos)

                if not hasattr(self, 'mouse_pos'):
                    self.mouse_pos = pos
                    return {'RUNNING_MODAL'}

                pos = pos - self.mouse_pos + self.obj_pos
                delta = Vector(mouse_pos) - self.oldpos if hasattr(self, 'oldpos') else Vector([0,0])
                self.oldpos = Vector([event.mouse_region_x, event.mouse_region_y])

                scene = bpy.context.scene
                saved_gravity = scene.gravity.copy()

                force = pos - object.location

                if event.ctrl:
                    for obj in self.get_selected_objects():
                        if obj != object:
                            obj.rigid_body.type = 'PASSIVE'
                else:
                    for obj in self.get_selected_objects():
                        obj.rigid_body.type = 'ACTIVE'

                if event.alt:
                    bpy.context.scene.gravity = Vector((0,0,1)) * (force * bpy.context.scene.render.fps * 30.0)

                elif event.shift:
                    if hasattr(self, 'mouse_pos'):
                        delattr(self, 'mouse_pos')

                    bpy.context.scene.gravity = Vector((0,0,0))

                    center = Vector((0,0,0))
                    n = 0
                    for obj in self.get_selected_objects():
                        center = center + obj.location
                        n = n + 1
                    if n:
                        center = center / n
                        self.charge_obj.location = center
                    
                    self.charge_obj.field.strength = delta.x * bpy.context.scene.render.fps
                else:
                    bpy.context.scene.gravity = force * bpy.context.scene.render.fps * 30.0
                    self.charge_obj.field.strength = 0

                fr = scene.frame_current

                scene.frame_set(fr + 1)

                mat_by_obj = dict()
                # d = bpy.context.evaluated_depsgraph_get()
                # d.update()
                for obj in self.get_selected_objects():
                    # object_eval = obj.evaluated_get(d)
                    # mat_by_obj[obj] = object_eval.matrix_world.copy()
                    mat_by_obj[obj] = obj.matrix_world.copy()

                scene.frame_set(fr)
                for obj, mat in mat_by_obj.items():
                    obj.matrix_basis = mat

                bpy.context.scene.gravity = saved_gravity

                if event.shift:
                    self.obj_pos = object.location.copy()


            except Exception as e:
                bpy.ops.wm.dialog('INVOKE_DEFAULT',
                    message=f'{e}',
                    title='Error during operation',
                    icon='ERROR')

                import traceback
                # print(e.__traceback__)
                print(traceback.format_exc())

                self.end(context, True)
                return {'FINISHED'}

        return {'RUNNING_MODAL'}


    def execute(self, context):
        try:
            if not len(bpy.context.selected_objects):
                return {'FINISHED'}
            
            wm = context.window_manager
            # self._timer = wm.event_timer_add(time_step=1/2, window=context.window)
            context.window_manager.modal_handler_add(self)
            self.init()
        
        except Exception as e:
            bpy.ops.wm.dialog('INVOKE_DEFAULT',
                message=f'{e}',
                title='Error during operation',
                icon='ERROR')

            import traceback
            # print(e.__traceback__)
            print(traceback.format_exc())

            return {'FINISHED'}
        
        return {'RUNNING_MODAL'}



################################################################
# Register 
addon_keymaps = []
classes = [
    PHYXDRAG_OT_simulate,
]


def register():
    tools.register(classes)

    tools.add_keymap(addon_keymaps, PHYXDRAG_OT_simulate.bl_idname, 'F', name='Object Mode')



def unregister():
    tools.clean_keymap(addon_keymaps)

    tools.unregister(classes)
